import functools
import inspect
import itertools
import math
import operator
import types
from typing import Dict
from typing import List

import numpy as np
import torch

from torchdynamo.guards import GuardBuilder
from torchdynamo.variables.tensor import DynamicShapeVariable

from .. import config
from .. import variables as vars
from ..allowed_functions import is_allowed
from ..exc import Unsupported
from ..exc import unimplemented
from ..source import AttrSource
from ..source import TypeSource
from ..utils import check_constant_args
from ..utils import istype
from ..utils import proxy_args_kwargs
from ..variables import Variable
from .base import MutableLocal


class BuiltinVariable(Variable):
    @staticmethod
    @functools.lru_cache(None)
    def _constant_fold_functions():
        fns = {
            abs,
            all,
            any,
            bool,
            callable,
            chr,
            dict,
            divmod,
            float,
            int,
            len,
            list,
            max,
            min,
            ord,
            pow,
            repr,
            round,
            set,
            str,
            str.format,
            sum,
            tuple,
            type,
            operator.pos,
            operator.neg,
            operator.not_,
            operator.invert,
            operator.pow,
            operator.mul,
            operator.matmul,
            operator.floordiv,
            operator.truediv,
            operator.mod,
            operator.add,
            operator.sub,
            operator.getitem,
            operator.lshift,
            operator.rshift,
            operator.and_,
            operator.or_,
            operator.xor,
            operator.ipow,
            operator.imul,
            operator.imatmul,
            operator.ifloordiv,
            operator.itruediv,
            operator.imod,
            operator.iadd,
            operator.isub,
            operator.ilshift,
            operator.irshift,
            operator.iand,
            operator.ixor,
            operator.ior,
            operator.index,
        }
        fns.update(x for x in math.__dict__.values() if isinstance(x, type(math.sqrt)))
        return fns

    def can_constant_fold_through(self):
        return self.fn in self._constant_fold_functions()

    @staticmethod
    @functools.lru_cache(None)
    def _fx_graph_functions():
        fns = {
            operator.pos,
            operator.neg,
            operator.not_,
            operator.invert,
            operator.pow,
            operator.mul,
            operator.matmul,
            operator.floordiv,
            operator.truediv,
            operator.mod,
            operator.add,
            operator.sub,
            operator.getitem,
            operator.lshift,
            operator.rshift,
            operator.and_,
            operator.or_,
            operator.xor,
            operator.ipow,
            operator.imul,
            operator.imatmul,
            operator.ifloordiv,
            operator.itruediv,
            operator.imod,
            operator.iadd,
            operator.isub,
            operator.ilshift,
            operator.irshift,
            operator.iand,
            operator.ixor,
            operator.ior,
        }
        return fns

    def can_insert_in_graph(self):
        return self.fn in self._fx_graph_functions()

    def __init__(self, fn, **kwargs):
        super(BuiltinVariable, self).__init__(**kwargs)
        self.fn = fn

    def python_type(self):
        return type(self.fn)

    def as_python_constant(self):
        return self.fn

    def reconstruct(self, codegen):
        name = self.fn.__name__
        assert self.fn.__module__ == "builtins"
        assert name not in codegen.tx.f_globals, "shadowed global"
        return [codegen.create_load_global(name, add=True)]

    def constant_args(self, *args, **kwargs):
        return check_constant_args(args, kwargs)

    def tensor_args(self, *args, **kwargs):
        return any(
            isinstance(i, vars.TensorVariable)
            for i in itertools.chain(args, kwargs.values())
        )

    def unspec_numpy_args(self, *args, **kwargs):
        return all(
            isinstance(
                i,
                (
                    vars.UnspecializedNumpyVariable,
                    vars.UnspecializedPythonVariable,
                    vars.ConstantVariable,
                ),
            )
            for i in itertools.chain(args, kwargs.values())
        ) and any(
            isinstance(x, vars.UnspecializedNumpyVariable)
            for x in itertools.chain(args, kwargs.values())
        )

    def unspec_python_args(self, *args, **kwargs):
        return all(
            isinstance(
                i,
                (
                    vars.UnspecializedPythonVariable,
                    vars.ConstantVariable,
                ),
            )
            for i in itertools.chain(args, kwargs.values())
        ) and any(
            isinstance(x, vars.UnspecializedPythonVariable)
            for x in itertools.chain(args, kwargs.values())
        )

    @staticmethod
    def unwrap_unspec_args_kwargs(args, kwargs):
        unwrapped_args = []
        unwrapped_kwargs = {}
        for x in args:
            if isinstance(
                x,
                (
                    vars.UnspecializedNumpyVariable,
                    vars.UnspecializedPythonVariable,
                ),
            ):
                unwrapped_args.append(x.raw_value)
            else:
                unwrapped_args.append(x.as_python_constant())
        for k, v in kwargs:
            if isinstance(
                x,
                (
                    vars.UnspecializedNumpyVariable,
                    vars.UnspecializedPythonVariable,
                ),
            ):
                unwrapped_kwargs.update({k: v.raw_value})
            else:
                unwrapped_kwargs.update({k: v.as_python_constant()})
        return unwrapped_args, unwrapped_kwargs

    def call_function(
        self, tx, args: List[Variable], kwargs: Dict[str, Variable]
    ) -> Variable:
        constant_args = check_constant_args(args, kwargs)
        tensor_args = self.tensor_args(*args, **kwargs)
        options = vars.propagate(self, args, kwargs.values())
        has_constant_handler = self.can_constant_fold_through() and constant_args
        assert isinstance(args, list)
        assert isinstance(kwargs, dict)

        if (
            self.fn is operator.getitem
            and len(args) == 2
            and isinstance(args[1], vars.TensorVariable)
            and args[1].dtype == torch.bool
            and not config.dynamic_shapes
        ):
            unimplemented("dynamic Tensor.__getitem__(bool[])")

        if self.can_insert_in_graph() and tensor_args:
            try:
                fn = self.fn
                if self.fn is operator.iadd and isinstance(
                    args[0], vars.ConstantVariable
                ):
                    # Work around weird bug in hf_T5
                    fn, args = operator.add, [args[1], args[0]]
                proxy = tx.output.create_proxy(
                    "call_function",
                    fn,
                    *proxy_args_kwargs(args, kwargs),
                )
                if self.unspec_numpy_args(*args, **kwargs):
                    _args, _kwargs = self.unwrap_unspec_args_kwargs(args, kwargs)
                    raw_value = self.fn(*_args, **_kwargs)
                    return vars.UnspecializedNumpyVariable.create(
                        tx,
                        proxy,
                        raw_value=raw_value,
                        **options,
                    )
                elif self.unspec_python_args(*args, **kwargs):
                    _args, _kwargs = self.unwrap_unspec_args_kwargs(args, kwargs)
                    raw_value = self.fn(*_args, **_kwargs)
                    need_unwrap = any(
                        x.need_unwrap
                        for x in itertools.chain(args, kwargs.values())
                        if isinstance(x, vars.UnspecializedPythonVariable)
                    )
                    return vars.UnspecializedPythonVariable.create(
                        tx,
                        proxy,
                        raw_value=raw_value,
                        need_unwrap=need_unwrap,
                        **options,
                    )
                else:
                    return vars.TensorVariable.create(tx, proxy, **options)

            except NotImplementedError:
                unimplemented(f"partial tensor op: {self} {args} {kwargs}")

        # Handle cases like int(torch.seed())
        if self.fn is int and isinstance(args[0], DynamicShapeVariable):
            return args[0]

        handler = getattr(self, f"call_{self.fn.__name__}", None)

        if handler:
            try:
                inspect.signature(handler).bind(tx, *args, **kwargs)
            except TypeError as exc:
                if config.debug:
                    print("WARN: incorrect arg count", handler, exc)
                handler = None

        if handler:
            try:
                result = handler(tx, *args, **kwargs)
                if result is not None:
                    return result.add_options(options)
            except Unsupported as exc:
                if not has_constant_handler:
                    raise
                # Actually, we will handle this just fine
                exc.remove_from_stats()

        if has_constant_handler:
            # constant fold
            return vars.ConstantVariable(
                self.as_python_constant()(
                    *[x.as_python_constant() for x in args],
                    **{k: v.as_python_constant() for k, v in kwargs.items()},
                ),
                **options,
            )

        return super().call_function(tx, args, kwargs)

    def _call_min_max(self, tx, a, b):
        if self.tensor_args(a, b):
            if not isinstance(a, vars.TensorVariable):
                a, b = b, a
            assert isinstance(a, vars.TensorVariable)

            # convert min/max to torch ops
            if b.is_python_constant():
                kwargs = {"min": b} if (self.fn is max) else {"max": b}
                result = vars.torch(torch.clamp).call_function(tx, [a], kwargs)
            else:
                fn = {max: torch.maximum, min: torch.minimum}[self.fn]
                result = vars.torch(fn).call_function(tx, [a, b], {})

            # return unspec if both a, b are unspec or const
            if all(
                isinstance(
                    i,
                    (
                        vars.UnspecializedNumpyVariable,
                        vars.UnspecializedPythonVariable,
                        vars.ConstantVariable,
                    ),
                )
                for i in [a, b]
            ):
                if b.is_python_constant():
                    raw_b = b.as_python_constant()
                else:
                    raw_b = b.raw_value
                if self.fn is max:
                    raw_res = max(a.raw_value, raw_b)
                else:
                    raw_res = min(a.raw_value, raw_b)

                if isinstance(raw_res, np.number):
                    return vars.UnspecializedNumpyVariable.from_tensor_variable(
                        result, raw_res
                    )
                else:
                    need_unwrap = any(
                        x.need_unwrap
                        for x in [a, b]
                        if isinstance(x, vars.UnspecializedPythonVariable)
                    )
                    return vars.UnspecializedPythonVariable.from_tensor_variable(
                        result, raw_res, need_unwrap
                    )
            # otherwise return tensor
            else:
                return result
        else:
            unimplemented(f"unsupported min / max over args {str(a)}, {str(b)}")

    call_min = _call_min_max
    call_max = _call_min_max

    def call_range(self, tx, *args, **kwargs):
        if self.constant_args(*args, **kwargs):
            return vars.RangeVariable(
                value=range(
                    *[x.value for x in args],
                    **{k: v.value for k, v in kwargs.items()},
                ),
            )

    def call_slice(self, tx, *args):
        return vars.SliceVariable(args)

    def _call_iter_tuple_list(self, tx, obj=None):
        cls = vars.BaseListVariable.cls_for(self.fn)
        if obj is None:
            return cls(
                [],
                mutable_local=MutableLocal(),
            )
        elif obj.has_unpack_var_sequence(tx):
            guards = set()
            if obj.source:
                guards.add(obj.source.create_guard(GuardBuilder.LIST_LENGTH))
            return cls(
                list(obj.unpack_var_sequence(tx)),
                mutable_local=MutableLocal(),
                guards=guards,
            ).trace(self, obj)

    call_iter = _call_iter_tuple_list
    call_tuple = _call_iter_tuple_list
    call_list = _call_iter_tuple_list

    def call_zip(self, tx, *args):
        options = vars.propagate(self, args)
        if all(x.has_unpack_var_sequence(tx) for x in args):
            items = [
                vars.basetuple(list(item), **options)
                for item in zip(*[arg.unpack_var_sequence(tx) for arg in args])
            ]
            return vars.basetuple(items, **options)

    def call_enumerate(self, tx, *args):
        options = vars.propagate(self, args)
        if len(args) == 1:
            start = 0
        else:
            assert len(args) == 2
            assert isinstance(args[1], vars.ConstantVariable)
            start = args[1].as_python_constant()
        if args[0].has_unpack_var_sequence(tx):
            items = [
                vars.basetuple(
                    [vars.constant(idx, **options), var],
                    **options,
                )
                for idx, var in enumerate(args[0].unpack_var_sequence(tx), start)
            ]
            return vars.basetuple(items, **options)

    def call_mul(self, tx, a, b):
        if isinstance(a, (vars.ListVariable, vars.TupleVariable)) and isinstance(
            b, vars.ConstantVariable
        ):
            return a.__class__(
                items=a.items * b.as_python_constant(), mutable_local=MutableLocal()
            ).trace(self, a, b)
        elif isinstance(b, (vars.ListVariable, vars.TupleVariable)) and isinstance(
            a, vars.ConstantVariable
        ):
            return b.__class__(
                items=b.items * a.as_python_constant(), mutable_local=MutableLocal()
            ).trace(self, a, b)
        else:
            return a.call_method(tx, "__mul__", [b], {})

    def call_len(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__len__", args[1:], kwargs)

    def call_add(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__add__", args[1:], kwargs)

    def call_sub(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__sub__", args[1:], kwargs)

    def call_truediv(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__truediv__", args[1:], kwargs)

    def call_floordiv(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__floordiv__", args[1:], kwargs)

    def call_iadd(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__iadd__", args[1:], kwargs)

    def call_getitem(self, tx, *args, **kwargs):
        return args[0].call_method(tx, "__getitem__", args[1:], kwargs)

    def call_isinstance(self, tx, arg, isinstance_type):
        arg_type = arg.python_type()
        isinstance_type = isinstance_type.as_python_constant()

        if isinstance(arg, vars.TensorVariable) and arg.dtype is not None:
            return vars.ConstantVariable(arg.call_isinstance(isinstance_type))
        # UserDefinedObject with C extensions can have torch.Tensor attributes,
        # so break graph.
        if isinstance(arg, vars.UserDefinedObjectVariable) and isinstance(
            arg.value, types.MemberDescriptorType
        ):
            unimplemented(
                f"isinstance called on UserDefinedClass {arg} {isinstance_type}"
            )
        try:
            val = issubclass(arg_type, isinstance_type)
        except TypeError:
            val = arg_type is isinstance_type
        return vars.ConstantVariable(val)

    def call_super(self, tx, a, b):
        return vars.SuperVariable(a, b)

    def call_next(self, tx, arg):
        if isinstance(arg, vars.ListIteratorVariable):
            val, next_iter = arg.next_variables()
            tx.replace_all(arg, next_iter)
            return val
        elif isinstance(arg, vars.BaseListVariable):
            return arg.items[0].add_options(self, arg)

    def call_hasattr(self, tx, obj, attr):
        if attr.is_python_constant():
            name = attr.as_python_constant()
            return obj.call_hasattr(tx, name).add_options(self, obj, attr)

    def call_map(self, tx, fn, seq):
        if seq.has_unpack_var_sequence(tx):
            items = [fn.call_function(tx, [x], {}) for x in seq.unpack_var_sequence(tx)]
            return vars.basetuple(items).add_options(self, fn, seq)

    def call_sum(self, tx, seq, **kwargs):
        # Special case for sum on tuple of floats and ints
        if (
            isinstance(seq, (vars.ListVariable, vars.TupleVariable))
            and all(
                [
                    isinstance(x, vars.ConstantVariable)
                    and isinstance(x.value, (int, float))
                    for x in seq.items
                ]
            )
            and not kwargs
        ):
            new_list = [x.value for x in seq.items]
            return vars.ConstantVariable(sum(new_list))
        if seq.has_unpack_var_sequence(tx):
            start = kwargs.pop("start", vars.ConstantVariable(0)).as_python_constant()
            assert not kwargs
            items = seq.unpack_var_sequence(tx)[start:]
            return BuiltinVariable(functools.reduce).call_function(
                tx,
                [
                    BuiltinVariable(operator.add),
                    vars.basetuple(items),
                    vars.constant(0).add_options(self, seq),
                ],
                {},
            )

    def call_reduce(self, tx, function, iterable, initializer=None):
        if iterable.has_unpack_var_sequence(tx):
            items = iterable.unpack_var_sequence(tx)
            if initializer is None:
                value, items = items[0], items[1:]
            else:
                value = initializer
            for element in items:
                value = function.call_function(tx, [value, element], {})
            return value

    def call_getattr(self, tx, obj: Variable, name_var: Variable, default=None):
        from . import GetAttrVariable
        from . import PythonModuleVariable
        from . import TorchVariable
        from . import UserFunctionVariable

        options = vars.propagate(self, obj, name_var)
        guards = options["guards"]
        name = name_var.as_python_constant()

        if not name_var.is_python_constant():
            unimplemented("non-const getattr() name")

        if tx.output.side_effects.is_attribute_mutation(obj):
            try:
                # re-read a pending side effect?
                return tx.output.side_effects.load_attr(obj, name).add_options(options)
            except KeyError:
                pass

        if default is not None:
            hasattr_var = self.call_hasattr(tx, obj, name_var)
            guards.update(hasattr_var.guards)
            assert hasattr_var.as_python_constant() in (True, False)
            if not hasattr_var.as_python_constant():
                return default.add_guards(guards)

        if obj.source:
            source = AttrSource(obj.source, name)
            options["source"] = source
        else:
            source = None

        if isinstance(obj, vars.NNModuleVariable):
            return obj.var_getattr(tx, name).add_options(options)
        elif isinstance(obj, vars.TensorVariable) and name == "grad":
            if source:
                example_value = obj.proxy.node.meta["example_value"].grad
                return vars.build(tx, source)(example_value).add_options(options)
            else:
                unimplemented("tensor grad")
        elif isinstance(
            obj,
            (
                vars.TensorVariable,
                vars.NamedTupleVariable,
                vars.ConstantVariable,
                vars.UserDefinedClassVariable,
                vars.UserDefinedObjectVariable,
            ),
        ):
            try:
                return (
                    obj.var_getattr(tx, name).clone(source=source).add_options(options)
                )
            except NotImplementedError:
                return GetAttrVariable(obj, name, **options)
        elif isinstance(obj, TorchVariable):
            member = getattr(obj.value, name)
            if is_allowed(member):
                return vars.torch(member, **options)
            elif vars.is_literal(member):
                return vars.constant(member, **options)
            else:
                return vars.build(tx, source)(member).add_guards(guards)
        elif isinstance(obj, PythonModuleVariable):
            member = obj.value.__dict__[name]
            return vars.build(tx, source)(member).add_guards(guards)
        elif istype(obj, UserFunctionVariable) and name in ("__name__", "__module__"):
            return vars.constant(getattr(obj.fn, name), **vars.propagate(obj))
        else:
            try:
                return (
                    obj.var_getattr(tx, name).clone(source=source).add_options(options)
                )
            except NotImplementedError:
                return GetAttrVariable(obj, name, **options)

    def call_setattr(self, tx, obj: Variable, name_var: Variable, val: Variable):
        if isinstance(obj, (vars.BlackHoleVariable, vars.DataClassVariable)):
            return obj.call_method(tx, "__setattr__", [name_var, val], {})
        elif (
            tx.output.side_effects.is_attribute_mutation(obj)
            and name_var.is_python_constant()
        ):
            tx.output.side_effects.store_attr(obj, name_var.as_python_constant(), val)
            return val.add_options(self, obj, name_var)
        elif isinstance(obj, vars.UserDefinedObjectVariable):
            unimplemented(
                f"setattr(UserDefinedObjectVariable) {type(obj.value).__setattr__}"
            )
        elif isinstance(obj, vars.NNModuleVariable):
            obj.convert_to_unspecialized(tx)

    def call_type(self, tx, obj: Variable):
        try:
            py_type = obj.python_type()
        except NotImplementedError:
            py_type = None

        if istype(obj, vars.TupleVariable):
            return BuiltinVariable(py_type).add_options(self, obj)

        if py_type is not None and obj.source:
            return vars.build(tx, TypeSource(obj.source))(py_type).add_options(
                self, obj
            )

        unimplemented(f"type({obj})")

    def call_reversed(self, tx, obj: Variable):
        if obj.has_unpack_var_sequence(tx):
            items = list(reversed(obj.unpack_var_sequence(tx)))
            return vars.basetuple(items, **vars.propagate(self, obj))

    def call_chain(self, tx, *args):
        if all(obj.has_unpack_var_sequence(tx) for obj in args):
            items = []
            for obj in args:
                items.extend(obj.unpack_var_sequence(tx))
            return vars.basetuple(items, **vars.propagate(self, *args))

    def call_islice(self, tx, iterable, *args):
        if iterable.has_unpack_var_sequence(tx) and all(
            x.is_python_constant() for x in args
        ):
            const_args = [x.as_python_constant() for x in args]
            items = iterable.unpack_var_sequence(tx)
            items = list(itertools.islice(items, *const_args))
            return vars.basetuple(items, **vars.propagate(self, iterable, *args))

    def call_id(self, tx, *args):
        if len(args) > 0 and isinstance(args[0], vars.NNModuleVariable):
            nn_mod_variable = args[0]
            mod = tx.output.get_submodule(nn_mod_variable.module_key)
            return vars.constant(id(mod))
        else:
            unimplemented(f"call_id with args {args}")
